#!/usr/bin/python
# -*- coding: latin-1 -*-

""" Author : DILDA Nathan 
Student at the Ecole Supérieur de Biotechnologie de Strasbourg (ESBS), France 
"""

from Tkinter import *
from ttk import Progressbar
import re
from itertools import product
from time import time
from tkFileDialog import askopenfilename, asksaveasfile

running=True
rapide=0

aa=['Alanine', 'Arginine', 'Asparagine', 'Aspartic acid', 'Cysteine',
    'Glutamic acid', 'Glutamine', 'Glycine', 'Histidine', 'Isoleucine',
    'Leucine', 'Lysine', 'Methionine', 'Phenylalanine','Proline',
    'Serine', 'Threonine', 'Tryptophan', 'Tyrosine', 'Valine','Stop']
    
alphaa={'Alanine':'A', 'Arginine':'R', 'Asparagine':'N', 'Aspartic acid':'D', 'Cysteine':'C',
    'Glutamic acid':'E', 'Glutamine':'Q', 'Glycine':'G', 'Histidine':'H', 'Isoleucine':'I',
    'Leucine':'L', 'Lysine':'K', 'Methionine':'M', 'Phenylalanine':'F','Proline':'P',
    'Serine':'S', 'Threonine':'T', 'Tryptophan':'W', 'Tyrosine':'Y', 'Valine':'V', 'Stop':'*'}

AA={'TAA':'*','TAG':'*','TGA':'*','CTT': 'L', 'ATG': 'M', 'ACA': 'T', 'ACG': 'T', 'ATC': 'I',
        'AAC': 'N', 'ATA': 'I', 'AGG': 'R', 'CCT': 'P', 'ACT': 'T', 'AGC': 'S', 'AAG': 'K', 'AGA': 'R',
        'CAT': 'H', 'AAT': 'N', 'ATT': 'I', 'CTG': 'L', 'CTA': 'L', 'CTC': 'L', 'CAC': 'H', 'AAA': 'K',
        'CCG': 'P', 'AGT': 'S', 'CCA': 'P', 'CAA': 'Q', 'CCC': 'P', 'TAT': 'Y', 'GGT': 'G', 'TGT': 'C',
        'CGA': 'R', 'CAG': 'Q', 'TCT': 'S', 'GAT': 'D', 'CGG': 'R', 'TTT': 'F', 'TGC': 'C', 'GGG': 'G',
        'GGA': 'G', 'TGG': 'W', 'GGC': 'G', 'TAC': 'Y', 'TTC': 'F', 'TCG': 'S', 'TTA': 'L', 'TTG': 'L',
        'TCC': 'S', 'ACC': 'T', 'TCA': 'S', 'GCA': 'A', 'GTA': 'V', 'GCC': 'A', 'GTC': 'V', 'GCG': 'A',
        'GTG': 'V', 'GAG': 'E', 'GTT': 'V', 'GCT': 'A', 'GAC': 'D', 'CGT': 'R', 'GAA': 'E', 'CGC': 'R',
        'TAA' : '*', 'TAG': '*', 'TGA': '*'}

def checkdna(event):

    global AA
    
    # Function which is triggered for eahc new letter written in the sequence text box
    # In checks that the sequence is a DNA sequence ( A / T / C / G )
    # It converts the sequence in uppercase

    # It also convert the DNA sequence in Protein equivalent and print this protein sequence in the dedicated zone

    typed_seq = Seq.get("1.0", END).replace('\n','').upper() # Sequence actuellement crite convertie en majuscule sans les sauts de lignes

    a=re.compile('^[ATCGatcg]*$')
    sprot=""
    currentpos=Seq.index(INSERT)
  
    if a.match(typed_seq) is None: # Vérifie que la séquence est bien de l'ADN
        Seq.configure(background="#ffacac")
        Boutonvalider.config(text='ERREUR')
        Boutonvalider.config(state=DISABLED)
    else:
        Seq.config(bg="white")
        Boutonvalider.config(state=NORMAL)
        Boutonvalider.config(text='Valider')
        
    # La squence est convertie en majuscule
    Seq.delete("1.0", END)
    Seq.insert("1.0", typed_seq)

    liste_codons = [typed_seq[i:i+3] for i in range(0,len(typed_seq),3)] # Liste des codons de la chaine
    for i in liste_codons: # Création de la séquence protéique
        if len(i)==3 and i.upper() in AA.keys():
            sprot=sprot+AA[i.upper()]
            
    seqprot.config(state="normal") # On active la zone de texte de la protine pour ajouter la sequence puis on reverouille
    seqprot.delete("1.0",END)
    seqprot.insert("1.0",sprot)
    seqprot.config(state=DISABLED)
    Seq.mark_set(INSERT, currentpos) # On remet le curseur sur la zone de la sequence ADN
    
def selaa(event):

    # Function that allows to select text in the protein sequence widget
    
    Seq.tag_remove("start", "1.0", END) # Removes all selection in both DNA and protein sequence
    seqprot.tag_remove("start", "1.0", END)
    
    if seqprot.tag_ranges("sel"): # If selection in protein sequence colors the selection
        seqprot.tag_add("start", "sel.first", "sel.last")
        last_index = int(seqprot.index('sel.last')[2:])
        first_index = int(seqprot.index('sel.first')[2:])

        if len(seqprot.get("sel.first", "sel.last"))==1: # Colors the corresponding codon in DNA sequence
            Seq.tag_add("start", "1."+str(first_index*3), "1."+str(last_index*3))
            Seq.tag_add(SEL, "1."+str(first_index*3), "1."+str(last_index*3))
           
    
def selseq(event):

    # Function that allows to select text in the DNA sequence widget

    seqprot.tag_remove("start", "1.0", END) # Removes all selection in both DNA and protein sequence
    Seq.tag_remove("start", "1.0", END)
    
    if Seq.tag_ranges("sel"): # If selection in DNA sequence colors the selection
        Seq.tag_add("start", "sel.first", "sel.last")
        last_index = int(Seq.index('sel.last')[2:])
        first_index = int(Seq.index('sel.first')[2:])

        if first_index%3==0 and len(Seq.get("sel.first", "sel.last"))==3: # If codon selected, select the corresponding amino acid in protein sequence
            seqprot.tag_add("start", "1."+str(last_index/3-1), "1."+str(last_index/3))
            

def selectall(event):
    Seq.tag_add(SEL, "1.0", END)
    return "break"

def prodcart(ar_list):
    if not ar_list:
        yield ()
    else:
        for a in ar_list[0]:
            for prod in prodcart(ar_list[1:]):
                yield (a,)+prod
                
def compseq(a,b):
    count=0
    for i in range(len(a)):
        if a[i]!=b[i]:
            count+=1
    return count    
    
def importer():
    a=re.compile('^[ATCGatcg]*$')
    im = askopenfilename(filetypes=[('FASTA file', '.fasta'), ('Text file', '.txt')])
    im = open(im,'r')
    for i in im:
        i=i.replace('\n','')
        if a.match(i):
            Seq.insert(END, i)
    checkdna('<KeyRelease>')
    
def exporter(id="0",comm="0"):

    if id=="0" and comm=="0":
        fenexport=Toplevel() 
        fenexport.resizable(width=False, height=False)
        fenexport.focus_set()
        fenexport.title('DMRS - Exportation de Séquence')
        lablid = Label(fenexport, text = 'Identifiant :').grid(row =0, column=0, pady=3, padx=2)
        id = Text(fenexport, bg ='white', fg='black', takefocus="", font="Courier 14", width=15, height=1, exportselection=0)
        id.grid(row=0,column=1)
        lablcomm = Label(fenexport, text = 'Commentaire:').grid(row =1, column=0, pady=3, padx=2)
        comm = Text(fenexport, bg ='white', fg='black', takefocus="", font="Courier 14", width=15, height=1, exportselection=0)
        comm.grid(row=1,column=1)
        Boutonvalider = Button(fenexport, text='Exporter', command= lambda: exporter(id,comm)).grid(row=2)
    
    elif id!="0" and comm!="0":
        dest = asksaveasfile(mode='w', defaultextension=".fasta")
    
        if not dest:
            return
        text2save = ">"+id.get("1.0", 'end-1c')+"| "+comm.get("1.0",'end-1c')+'\n'+Seq.get("1.0",'end-1c')
        dest.write(text2save)
        dest.close()
    
def Antisens():
    TRANS = { "T": "A", "A": "T", "G": "C", "C": "G" }
    Seqsens = Seq.get("1.0", END).replace('\n','')
    Seqanti = ''.join([TRANS[i] for i in Seqsens])[::-1]
    Seq.delete("1.0", END)
    Seq.insert("1.0", Seqanti)
    checkdna(0)
    
def expfile():

    ### Recursive ###
    dest = asksaveasfile(mode='w', defaultextension=".fasta")
    
    if not dest:
        return
    
    print Seq.get("1.0",END)
    print comm.get("1.0",END)
    text2save = ">"+identif.get("1.0", END)+"|"+comm.get("1.0",END)+'\n'+Seq.get("1.0",END)
    print text2save
    dest.write(text2save)
    dest.close()

def diffseq(seqM,seq):
    traits=''
    for i in range(len(seq)):
            if seqM[i]==seq[i]:
                traits+='|'
            else:
                traits+=' '
    return traits

def annuler():
	global running
	running=False
    
def Rapide():
    global rapide
    
    if rapide==0:
        rapide=1
        menuaction.entryconfig(1, label='Recherche normale')
    else:
        rapide=0
        menuaction.entryconfig(1, label='Recherche rapide')

    

def info():
    fen3=Toplevel() # Création de la nouvelle fenêtre
	
    # Obtention du focus pour la nouvelle fenêtre
    fen3.grab_set()
    fen3.focus_set()
    fen3.resizable(width=False, height=False)
    labinfotitre = Label(fen3, text="Mode d'emploi :\n-------------------------------------", font="Arial 14 bold").grid(row=0)
    labinfo = Label(fen3, text="1°) Entrez une séquence nucléotidiques\n2°)Sélectionnez avec le curseur l'acide aminé à modifier\n3°)Choisissez dans la liste le nouvel acide aminé\n4°)Cliquez sur valider\n__________________\n\nle programme détermine la nouvelle séquence\net l'enzyme de restriction\nqui permettra la vérification\n=====================").grid(row=1)
    labinfoenz = Label(fen3, text="L'onglet Fichier permet d'importer\nou d'exporter une séquence FASTA\n=====================\n\nL'onglet Action vous permet d'obtenir le brin complémentaire\nou d'activer la recherche rapide\nLa recherche rapide ignore certaines\ncombinaisons et est 3 fois plus rapide\nVous aurez cependant moins de choix").grid(row=2)
    labinfocop = Label(fen3, text="\nCopyright DILDA Nathan 2015\nÉtudiant à l'École Supérieure de Biotechnologies de Strasbourg", font="Arial 10").grid(row=3)

def mutation():
    
    global running
    global rapide
    global AA
    
    # Function will generate a list of every possible combinaisons of codons that introduce the mutation but keep all other amino acid
    # It will then check the presence of new restriction site
    
    AA={'TAA':'*','TAG':'*','TGA':'*','CTT': 'L', 'ATG': 'M', 'ACA': 'T', 'ACG': 'T', 'ATC': 'I', 'AAC': 'N', 'ATA': 'I', 'AGG': 'R', 'CCT': 'P', 'ACT': 'T', 'AGC': 'S', 'AAG': 'K', 'AGA': 'R', 'CAT': 'H', 'AAT': 'N', 'ATT': 'I', 'CTG': 'L', 'CTA': 'L', 'CTC': 'L', 'CAC': 'H', 'AAA': 'K', 'CCG': 'P', 'AGT': 'S', 'CCA': 'P', 'CAA': 'Q', 'CCC': 'P', 'TAT': 'Y', 'GGT': 'G', 'TGT': 'C', 'CGA': 'R', 'CAG': 'Q', 'TCT': 'S', 'GAT': 'D', 'CGG': 'R', 'TTT': 'F', 'TGC': 'C', 'GGG': 'G', 'GGA': 'G', 'TGG': 'W', 'GGC': 'G', 'TAC': 'Y', 'TTC': 'F', 'TCG': 'S', 'TTA': 'L', 'TTG': 'L', 'TCC': 'S', 'ACC': 'T', 'TCA': 'S', 'GCA': 'A', 'GTA': 'V', 'GCC': 'A', 'GTC': 'V', 'GCG': 'A', 'GTG': 'V', 'GAG': 'E', 'GTT': 'V', 'GCT': 'A', 'GAC': 'D', 'CGT': 'R', 'GAA': 'E', 'CGC': 'R'}
    
    # Get the DNA sequence and the amino acid to introduce
    seq=Seq.get("1.0",END).upper().replace('\n','')
    protseq = seqprot.get("1.0",END).replace('\n','')

    naa=alphaa[lbox.get(lbox.curselection())]

    indxseq=int(Seq.index('start.first')[2:])
    indxprotseq = int(seqprot.index('start.first')[2:])

    enzpresI=[] # Restriction site present in the initial sequence
    enzpresM={} # Restriction site in the new sequence

    # Get the list of Restriction Enzyme and their restriction site
    fichier_enz=open('fichier_enz_final','r')

    # Test the presence of restriction site in the initial sequence
    for lign in fichier_enz:
        restsite=re.compile(lign.split()[1])
        if (re.search(restsite,seq)) and (lign.split()[0] not in enzpresI):
            enzpresI.append(lign.split()[0])
    fichier_enz.close()
    
    if re.match('^[ATCG]*$',seq) is None: # Check if sequence is a DNA sequence otherwise raise Exception
        raise Exception("The Sequence is not a DNA sequence")

    if naa not in AA.values() or naa == '': # Check if the Amino acid to introduce exist, otherwise raise Exception
        raise Exception("This Amino Acid does not exist or you did not select an amino acid")

    Ltot = [] # List cointaining list of codons for each amino acid of the final protein sequence
    
    nseq=""
    enz=""

    aap = [protseq[i] for i in range(max(indxprotseq-3,0),indxprotseq)] # Get the 3 amino acid before the mutation or less if not enough
    codonpp=seq[:max(indxseq-9,0)] # Get the sequence before the mutation and the 3 amino acids before

    #codonp=[seq[i:i+3] for i in range(min(indxseq,9),indxseq,3)] 
    #codonpp=seq[:min(indxseq,9)] # Get the sequence before the mutation and the 3 codons before

    aas = [protseq[i] for i in range(indxprotseq+1,min(indxprotseq+4,len(protseq)))] # Get the 3 amino acid after the mutation or less if not enough
    codonss=seq[min(len(seq),indxseq+12):] # Get the sequence after the mutation and the 3 amino acids after

    AAtot = aap + [naa] + aas # Final sequence of protein after mutation
    
    for j in AAtot: # Convert final protein sequence in list of codon lists
        Lc = [k for k, v in AA.iteritems() if v == j]
        Ltot.append(Lc)

    comb=1.0

    for l in Ltot: # Calcul du nombre de combinaisons total
        comb=comb*len(l)
    combencours=0.0
    
    if rapide==1:
        comb=comb/3

    timeun=time()

    for seqmute in prodcart(Ltot): # Test de toutes les squences possibles
    
        Boutonvalider.config(text='Annuler', command=annuler)
        
        if running==False:
	        break
        if rapide == 1 and combencours>=comb:
            break
            
        combencours+=1
        seqmute=codonpp+''.join(seqmute)+codonss
        fichier_enz=open('fichier_enz_final','r')

        for lign in fichier_enz: # Test de toutes les enzymes pour la séquence en cours
            restsite=re.compile(lign.split()[1])
            
            if (re.search(restsite,seqmute)) and (lign.split()[0] not in enzpresI):
    
                # On teste que l'enzyme en cours ne coupe pas déjà initialement la séquence
                
                if (tuple(lign.split()) in enzpresM) and compseq(seqmute,seq)<compseq(enzpresM[tuple(lign.split())],seq):
    
                    # Si on trouve un nouveau site de restriction pour cette enzyme, 
                    #on vérifie que les modifications sont moins importante dans la séquence
                    enzpresM[(lign.split()[0],lign.split()[1])]=seqmute
                    
                elif tuple(lign.split()) not in enzpresM:
    
                    enzpresM[(lign.split()[0],lign.split()[1])]=seqmute

        Fen1.title('DMRS - ' + 'Calcul en Cours: ' + str(float(combencours/comb)*100)[:4] + '% - ' + str(comb-combencours) + ' Combinaisons à tester...')
        Fen1.update()
        fichier_enz.close()
    
    nommutation = str(seqprot.get(seqprot.index('sel.first'),seqprot.index('sel.last'))+str(int(seqprot.index('sel.first')[2:])+1)+naa)
    Fen1.title('DMRS - Calcul Terminé - ' + nommutation)
    enzM =[]
    
    for i in enzpresM:
        enzM.append((i[0],enzpresM[i],diffseq(enzpresM[i],seq),seq,i[1]))
    print_result(enzM,timeun,indxseq,seq)
    Boutonvalider.config(text='Valider', command=mutation)
    running=True
    
def print_result(enzM,timeun,indxseq,seq):
    
    # Fonction permettant l'affichage des nouvelles séquences avec le site de restriction
    # Les différentes codnitions évitent des dépassements d'index si l'acide aminé sélectionné 
    # est trop proche du début ou de la fin
    
    newsequences=Text(Fen1, bg='#f9f8f8', fg='black', font='Courier 12', width=55, height=12)
    newsequences.tag_config("seqenz", font='Courier 12 bold', foreground='blue')
    newsequences.grid(row=2, column=0, columnspan=3, pady=5, padx=4)
    newsequences.tag_config("seqenz", font='Courier 12 bold', foreground='blue')
    count=1
    for proposition in enzM:
        indexcourant = newsequences.index(END)
        newsequences.insert(END, "Proposition "+str(count)+": "+proposition[0]+"("+proposition[4]+")\n\n")
        if indxseq<15 and len(seq)-indxseq>=18:
            newsequences.insert(END,"      "+proposition[1][:indxseq+18]+'...\n      ')
            countVar = StringVar()
            indexcourant=newsequences.search(proposition[4], indexcourant, regexp=True, count=countVar)
            newsequences.insert(END,proposition[2][:indxseq+18]+'\n      ')
            newsequences.insert(END,proposition[3][:indxseq+18]+'...\n\n')
        elif indxseq>15 and len(seq)-indxseq>=18:
            newsequences.insert(END,"      "+proposition[1][indxseq-15:indxseq+18]+'...\n      ')
            countVar = StringVar()
            indexcourant=newsequences.search(proposition[4], indexcourant, regexp=True, count=countVar)
            newsequences.insert(END,proposition[2][indxseq-15:indxseq+18]+'\n      ')
            newsequences.insert(END,proposition[3][indxseq-15:indxseq+18]+'...\n\n')
        elif indxseq>15 and len(seq)-indxseq<18:
            newsequences.insert(END,"      "+proposition[1][indxseq-15:]+'...\n      ')
            countVar = StringVar()
            indexcourant=newsequences.search(proposition[4], indexcourant, regexp=True, count=countVar)
            newsequences.insert(END,proposition[2][indxseq-15:]+'\n      ')
            newsequences.insert(END,proposition[3][indxseq-15:]+'...\n\n')
        else:
            newsequences.insert(END,"      "+proposition[1]+'...\n      ')
            countVar = StringVar()
            indexcourant=newsequences.search(proposition[4], indexcourant, regexp=True, count=countVar)
            newsequences.insert(END,proposition[2]+'\n      ')
            newsequences.insert(END,proposition[3]+'...\n\n')
            
        newsequences.tag_add("seqenz", indexcourant, indexcourant.split(".")[0]+"."+str(int(indexcourant.split(".")[1])+int(countVar.get())))
            
        
        count+=1
    newsequences.insert(END, "\n\n"+str(int((time()-timeun)/60))+' min '+str((time()-timeun)%60)+' sec')
    
# Création de la fenêtre principale
Fen1 = Tk()
Fen1.title('DMRS')
Fen1.resizable(width=False, height=False)

# Création frames

Framseq = LabelFrame(Fen1, text="Séquence:", padx=5, pady=5)
Framseq.grid(row=0, column=0,padx=3, pady=5, sticky=W)
Framaa=LabelFrame(Fen1, text="Acide Aminé :", padx=2, pady=4)
Framaa.grid(row=0,column=1,padx=1, pady=5, sticky=W)


# Création Text widget

Seq = Text(Framseq, bg ='white', fg='black', takefocus="", font="Courier 14", width=36, height=8, exportselection=0)
Seq.grid(padx=4, pady=2, sticky='w')

# Création Listbox widget

scro = Scrollbar(Framaa)
lbox = Listbox(Framaa, height=11, exportselection=0, width=18)

# Configuration et coloration Listbox

for i in range(1,len(aa)):
    lbox.insert(END, aa[i])
    if i%2==0: #Une ligne sur deux est grise
        lbox.itemconfig(i-1, bg="#dedede")

scro.config(command=lbox.yview)
lbox.config(yscrollcommand=scro.set)
lbox.grid(row=0,column=0,padx=2, pady=3, sticky=W)
scro.grid(row=0, column=1, columnspan=2,sticky='ns',rowspan=3)
lbox.select_set(0) #Select by default the first value
# Boutons

Boutonvalider = Button(Fen1, text='Valider', command=mutation)
Boutonvalider.grid(row=1,column=1,padx=2, pady=2,sticky=E)

# Configuration sequence et vnements

Seq.bind('<ButtonRelease-1>', selseq)
Seq.bind('<KeyRelease>', checkdna)
Seq.tag_config("start", font='Courier 14 bold', foreground='black', background='#3399ff')
Seq.bind('<Control-a>', selectall)
Framseqprot= Frame(Fen1).grid(row=1)
seqprot= Text(Framseqprot,width=50,height=2,state=DISABLED, font="Courier 12", background="#dedede")
seqprot.grid(row=1,pady=3,columnspan=2,sticky=W,padx=5)
seqprot.bind('<ButtonRelease-1>', selaa)

seqprot.tag_config("start",font='Courier 14 bold', foreground='black', background='#3399ff')

# Insertion menu 

menubar = Menu(Fen1)

    # Menufichier

menufichier = Menu(menubar, tearoff=0)
menubar.add_cascade(label="Fichier", menu=menufichier)
menufichier.add_command(label="Importer une séquence", command=importer)
menufichier.add_command(label="Exporter une séquence", command=exporter)
menufichier.add_separator()
menufichier.add_command(label="Quitter", command=Fen1.quit)

    # Menu Action
menuaction = Menu(menubar, tearoff=0)
menubar.add_cascade(label="Action",menu=menuaction)
menuaction.add_command(label="Complémentaire",command=Antisens)
menuaction.add_command(label="Recherche rapide",command=Rapide)

    # Menu Info
menuinfo = Menu(menubar, tearoff=0)
menubar.add_cascade(label="?",menu=menuinfo)
menuinfo.add_command(label="Info",command=info)

Fen1.config(menu=menubar)
Fen1.mainloop()
